@@ -1,15 +1,17 @@
+import uuid
 from pathlib import Path, PurePosixPath
 import datetime
 import logging
 from uuid import UUID
 import itertools
+from collections import defaultdict
 
 import requests.exceptions
 
 from iblutil.io import hashfile
 
 import one.alf.io as alfio
-from one.alf.files import rel_path_parts, session_path_parts, get_session_path
+from one.alf.files import rel_path_parts, session_path_parts, get_session_path, get_alf_path
 import one.alf.exceptions as alferr
 from one.api import ONE
 from one.util import ensure_list
@@ -174,9 +176,10 @@ class RegistrationClient:
         assert not self.one.offline, 'ONE must be in online mode'
         date = self.ensure_ISO8601(date)  # Format, validate
         # Ensure subject exists on Alyx
-        self.assert_subject_exists(subject)
-        session_root = Path(session_root or self.one.alyx.cache_dir) / subject / f'{date:%Y-%m-%d}'
-        session_path = alfio.next_num_folder(session_root)
+        self.assert_exists(subject, 'subjects')
+        session_root = Path(session_root or self.one.alyx.cache_dir) / subject / date[:10]
+        session_path = session_root / alfio.next_num_folder(session_root)
+        session_path.mkdir(exist_ok=True, parents=True)  # Ensure folder exists on disk
         eid = UUID(self.create_session(session_path)['url'][-36:]) if register else None
         return session_path, eid
 
@@ -199,74 +202,44 @@ class RegistrationClient:
         return (x for x in dsets if x.is_file() and
                 any(x.name.endswith(y) for y in self.file_extensions))
 
-    def assert_subject_exists(self, subject: str) -> None:
-        """Raise an error if a given subject doesn't exist on Alyx database
+    def assert_exists(self, member, endpoint):
+        """Raise an error if a given member doesn't exist on Alyx database
 
         Parameters
         ----------
-        subject : str
-            The subject nickname to verify
+        member : str, uuid.UUID, list
+            The member ID(s) to verify
+        endpoint: str
+            The endpoint at which to look it up
+
+        Examples
+        --------
+        client.assert_exists('ALK_036', 'subjects')
+        client.assert_exists('user_45', 'users')
+        client.assert_exists('local_server', 'repositories')
 
         Raises
         -------
         one.alf.exceptions.AlyxSubjectNotFound
             Subject does not exist on Alyx
-        requests.exceptions.HTTPError
-            Failed to connect to Alyx database
-        """
-        try:
-            subject = self.one.alyx.rest('subjects', 'read', id=subject, no_cache=True)
-        except requests.exceptions.HTTPError as ex:
-            if '404' not in str(ex):
-                raise ex
-            else:
-                raise alferr.AlyxSubjectNotFound(subject)
-
-    def assert_user_exists(self, user):
-        """Raise an error if a given user doesn't exist on Alyx database
-
-        Parameters
-        ----------
-        user : str, list
-            The username to verify
-
-        Raises
-        -------
         one.alf.exceptions.ALFError
-            User does not exist on Alyx
+            Member does not exist on Alyx
         requests.exceptions.HTTPError
-            Failed to connect to Alyx database
+            Failed to connect to Alyx database or endpoint not found
         """
-        if isinstance(user, str):
+        if isinstance(member, (str, uuid.UUID)):
             try:
-                self.one.alyx.rest('users', 'read', id=user, no_cache=True)
+                self.one.alyx.rest(endpoint, 'read', id=str(member), no_cache=True)
             except requests.exceptions.HTTPError as ex:
                 if '404' not in str(ex):
                     raise ex
+                elif endpoint == 'subjects':
+                    raise alferr.AlyxSubjectNotFound(member)
                 else:
-                    raise alferr.ALFError(f'User "{user}" doesn\'t exist in Alyx')
+                    raise alferr.ALFError(f'Member "{member}" doesn\'t exist in Alyx')
         else:
-            for x in user:
-                self.assert_user_exists(x)
-
-    def assert_repository_exists(self, name):
-        """
-
-        Parameters
-        ----------
-        name
-
-        Returns
-        -------
-
-        """
-        try:
-            self.one.alyx.rest('repository', 'read', id=name, no_cache=True)
-        except requests.exceptions.HTTPError as ex:
-            if '404' not in str(ex):
-                raise ex
-            else:
-                raise alferr.ALFError(f'Repository "{name}" doesn\'t exist in Alyx')
+            for x in member:
+                self.assert_exists(x, endpoint)
 
     @staticmethod
     def ensure_ISO8601(date) -> str:
@@ -326,7 +299,7 @@ class RegistrationClient:
             ses_path = Path(ses_path)
         details = session_path_parts(ses_path.as_posix(), as_dict=True, assert_valid=True)
         # query alyx endpoints for subject, error if not found
-        self.assert_subject_exists(details['subject'])
+        self.assert_exists(details['subject'], 'subjects')
 
         # look for a session from the same subject, same number on the same day
         session_id, session = self.one.search(subject=details['subject'],
@@ -334,14 +307,11 @@ class RegistrationClient:
                                               number=details['number'],
                                               details=True, query_type='remote')
         users = ensure_list(users or self.one.alyx.user)
-        self.assert_user_exists(users)
-        # this is the generic relative path: subject/yyyy-mm-dd/NNN
-        rel_path = PurePosixPath(details['subject'], details['date'], details['number'])
+        self.assert_exists(users, 'users')
 
         # if nothing found create a new session in Alyx
         ses_ = {'subject': details['subject'],
                 'users': users,
-                'lab': details['lab'] or None,
                 'type': 'Experiment',
                 'number': details['number']}
         if kwargs.get('end_time', False):
@@ -350,6 +320,8 @@ class RegistrationClient:
         if kwargs.get('procedures', False):
             ses_['procedures'] = ensure_list(kwargs.pop('procedures'))
         assert ('subject', 'number') not in kwargs
+        if 'lab' not in kwargs and details['lab']:
+            kwargs.update({'lab': details['lab']})
         ses_.update(kwargs)
 
         if not session:  # Create from scratch
@@ -362,39 +334,43 @@ class RegistrationClient:
 
         _logger.info(session['url'] + ' ')
         # at this point the session has been created. If create only, exit
-        if file_list: #
-            self.register_datasets()
+        if file_list:
+            self.register_files(self.find_files(ses_path) if file_list is True else file_list)
 
         return session
 
-    def register_files(self, filelist):
-        F = []  # empty list whose keys will be relative paths and content filenames
+    def register_files(self, file_list):
+        F = defaultdict(list)  # empty map whose keys will be session paths
         md5s = []
         file_sizes = []
+        file_list = list(file_list)
 
-        for fn in self.find_files(ses_path) if file_list is True else file_list:
-            if fn.suffix in EXCLUDED_EXTENSIONS:
-                _logger.debug('Excluded: ', str(fn))
-                continue
-            if not _check_filename_for_registration(fn, self.registration_patterns):
-                _logger.warning('No matching dataset type for: ' + str(fn))
+        # Filter valid files and sort by session
+        for fn in file_list:
+            session_path = get_session_path(fn)
+            if fn.suffix not in self.file_extensions:
+                _logger.debug(f'{fn}: No matching extension {fn.suffix} in database')
                 continue
             if fn.suffix not in self.file_extensions:
-                _logger.warning('No matching dataformat (ie. file extension) for: ' + str(fn))
+                _logger.warning('No matching data format (i.e. file extension) for: ' + str(fn))
                 continue
-            if not _register_bool(fn.name, file_list):
-                _logger.debug('Not in filelist: ' + str(fn))
-                continue
-            try:
-                assert (str(gen_rel_path) in str(fn))
-            except AssertionError as e:
-                strerr = 'ALF folder mismatch: data is in wrong subject/date/number folder. \n'
-                strerr += ' Expected ' + str(gen_rel_path) + ' actual was ' + str(fn)
-                _logger.error(strerr)
-                raise e
+            F[session_path].append(fn)
+
+        for session_path, files in F.items():
+            # this is the generic relative path: subject/yyyy-mm-dd/NNN
+            get_alf_path()
+            details = session_path_parts(session_path.as_posix(), as_dict=True, assert_valid=True)
+            rel_path = PurePosixPath(details['subject'], details['date'], details['number'])
+            # try:
+            #     assert (str(rel_path) in str(fn))
+            # except AssertionError as e:
+            #     strerr = 'ALF folder mismatch: data is in wrong subject/date/number folder. \n'
+            #     strerr += ' Expected ' + str(gen_rel_path) + ' actual was ' + str(fn)
+            #     _logger.error(strerr)
+            #     raise e
             # extract the relative path of the file
-            rel_path = Path(str(fn)[str(fn).find(str(gen_rel_path)):])
-            F.append(str(rel_path.relative_to(gen_rel_path).as_posix()))
+            rel_path = Path(str(fn)[str(fn).find(str(session_path)):])
+            F[session_path].append(str(rel_path.relative_to(session_path).as_posix()))
             file_sizes.append(fn.stat().st_size)
             md5s.append(hashfile.md5(fn) if fn.stat().st_size < 1024 ** 3 else None)
             _logger.info('Registering ' + str(fn))
@@ -445,10 +421,10 @@ class RegistrationClient:
             Failed to connect to database, or submitted data not valid (500)
         """
         # Ensure subject exists
-        self.assert_subject_exists(subject)
+        self.assert_exists(subject, 'subjects')
         # Ensure user(s) exist
         user = ensure_list(kwargs.pop('user', [])) or self.one.alyx.user
-        self.assert_user_exists(user)
+        self.assert_exists(user, 'users')
         # Ensure volume not zero
         if volume == 0:
             raise ValueError('Water volume must be greater than zero')
@@ -500,10 +476,10 @@ class RegistrationClient:
             Failed to connect to database, or submitted data not valid (500)
         """
         # Ensure subject exists
-        self.assert_subject_exists(subject)
+        self.assert_exists(subject, 'subjects')
         # Ensure user(s) exist
         user = user or self.one.alyx.user
-        self.assert_user_exists(user)
+        self.assert_exists(user, 'users')
         # Ensure weight not zero
         if weight == 0:
             raise ValueError('Water volume must be greater than 0')
